#
#   Cryptkit test program - procedures
#   
#   based on the test programs in CryptlibConverter
#   by Trevor Perrin <trevp at trevp.net> - http://trevp.net/cryptlibConverter/
#

lappend auto_path [pwd]/lib
package require cryptkit
namespace import cryptkit::*

proc trace {args} {
    puts stderr "[lindex $args 0] = [uplevel $args]"
}

proc openKeyStore {} {
    global keySet
    cryptKeysetOpen keySet CRYPT_UNUSED CRYPT_KEYSET_FILE "keyset.p15" \
                           CRYPT_KEYOPT_CREATE
}

proc createKeyPairAndCert {commonName label isCa isCodeSigning} {
    global keySet caKeyPair
    puts "Generating key pair and cert for $label"

    # Create key pair
    cryptCreateContext keyPair CRYPT_UNUSED CRYPT_ALGO_RSA
    cryptSetAttributeString $keyPair CRYPT_CTXINFO_LABEL $label
    cryptSetAttribute $keyPair CRYPT_CTXINFO_KEYSIZE 128
    cryptGenerateKey $keyPair
    if {$isCa} {
        set caKeyPair $keyPair
    }
    # Create cert
    cryptCreateCert cert CRYPT_UNUSED CRYPT_CERTTYPE_CERTIFICATE
    if {$isCa} {
        cryptSetAttribute $cert CRYPT_CERTINFO_SELFSIGNED 1
        cryptSetAttribute $cert CRYPT_CERTINFO_CA 1
    }
    if {$isCodeSigning} {
        cryptSetAttribute $cert CRYPT_CERTINFO_EXTKEY_CODESIGNING CRYPT_UNUSED
    }
    cryptSetAttribute $cert CRYPT_CERTINFO_SUBJECTPUBLICKEYINFO $keyPair
    cryptSetAttributeString $cert CRYPT_CERTINFO_COUNTRYNAME "us"
    cryptSetAttributeString $cert CRYPT_CERTINFO_COMMONNAME $commonName

    # Sign cert with keyPair
    cryptSignCert $cert $caKeyPair

    # Add keyPair and cert chain to pkcs15 keyset (.p15)
    cryptAddPrivateKey $keySet $keyPair "password"
    cryptAddPublicKey $keySet $cert

    # Write out cert chain to pkcs7 file (.p7b)
    cryptExportCert buffer bufferLength CRYPT_CERTFORMAT_CERTCHAIN $cert
    set fos [open $label.p7b w]
    puts $fos $buffer
    close $fos

    # Cleanup
    cryptDestroyContext $cert
    cryptDestroyContext $keyPair

    # Re-get the CA key pair so we get it attached to the certificate
    # This is a kludge, I wish I could just attach the cert directly
    # to the keypair, but we need to do this so we can later sign with
    # this keypair and have the cert automatically added to the chain
    if {$isCa} {
        cryptGetPrivateKey $keySet caKeyPair CRYPT_KEYID_NAME $label "password"
    }
}

proc closeKeyStore {} {
    global keySet caKeyPair
    cryptDestroyContext $caKeyPair
    cryptKeysetClose $keySet
}

proc makeTempKeyPairsAndCerts {closeKeySet} {
    global tempKeyPairs tempCerts tempKeySet
    puts "Generating temporary key pairs and certs"
    set tempKeyPairs [list]
    set tempCerts [list]
    cryptKeysetOpen tempKeySet CRYPT_UNUSED CRYPT_KEYSET_FILE "keyset.p15" \
                                CRYPT_KEYOPT_CREATE
    foreach count [list 0 1] {
        # Generate temp key pair
        cryptCreateContext k CRYPT_UNUSED CRYPT_ALGO_RSA
        cryptSetAttributeString $k CRYPT_CTXINFO_LABEL "Temp Key $count"
        cryptSetAttribute $k CRYPT_CTXINFO_KEYSIZE 128
        cryptGenerateKey $k
        #  Generate good-for-anything self-signed cert for key pair
        cryptCreateCert c CRYPT_UNUSED CRYPT_CERTTYPE_CERTIFICATE
        cryptSetAttribute $c CRYPT_CERTINFO_XYZZY 1 ;# special good-for-anything cert
        cryptSetAttribute $c CRYPT_CERTINFO_SUBJECTPUBLICKEYINFO $k
        cryptSetAttributeString $c CRYPT_CERTINFO_COMMONNAME "Temp Cert $count"
        cryptSignCert $c $k

        cryptAddPrivateKey $tempKeySet $k "password"
        cryptAddPublicKey $tempKeySet $c
        cryptDestroyContext $k
        cryptGetPrivateKey $tempKeySet k CRYPT_KEYID_NAME "Temp Key $count" \
                            "password"
        lappend tempKeyPairs $k
        lappend tempCerts $c
    }
    if {$closeKeySet} {
        cryptKeysetClose $tempKeySet
    }
}

proc destroyTempKeyPairsAndCerts {closeKeySet} {
    global tempKeyPairs tempCerts tempKeySet
    foreach context $tempKeyPairs {
        cryptDestroyContext $context
    }
    foreach cert $tempCerts {
        cryptDestroyCert $cert
    }
    if {$closeKeySet} {
        cryptKeysetClose $tempKeySet
    }
}

proc testTLSclient {serverName} {
    # Open an SSL session to the server
    cryptCreateSession session CRYPT_UNUSED CRYPT_SESSION_SSL
    cryptSetAttributeString $session CRYPT_SESSINFO_SERVER_NAME $serverName
    cryptSetAttribute $session CRYPT_SESSINFO_SERVER_PORT 2000
    cryptSetAttribute $session CRYPT_SESSINFO_ACTIVE 1
    # Set to blocking
    cryptSetAttribute CRYPT_UNUSED CRYPT_OPTION_NET_READTIMEOUT 30
    cryptSetAttribute CRYPT_UNUSED CRYPT_OPTION_NET_WRITETIMEOUT 30
    set text [binary format a10240 \0]
    set bytesSent 0
    set bytesRead 0
    set startTime [clock seconds]
    for {set count 0} {$count < 512} {incr count} {
        set bytesSent 0
        while {$bytesSent < 10240} {
            cryptPushData $session $text bytesCopied
            incr bytesSent $bytesCopied
        }
        cryptFlushData $session
        set bytesRead 0
        while {$bytesRead < 10240} {
            cryptPopData $session text 10240 bytesCopied
            incr bytesRead $bytesCopied
        }
    }
    set delay [expr {[clock seconds] - $startTime}]
    puts "Sent 10 MB over TLS 1.0 with 3DES in $delay seconds"
    cryptDestroySession $session
    puts "Session closed..."
}

proc testTLSserver {} {
    # Start up a single-threaded server, using the server keyPair/cert we
    # previously generated, and block waiting for a connection
    global tempKeyPairs tempCerts
    puts "Waiting for client..."
    cryptCreateSession session CRYPT_UNUSED CRYPT_SESSION_SSL_SERVER
    cryptSetAttributeString $session CRYPT_SESSINFO_SERVER_NAME localhost
    cryptSetAttribute $session CRYPT_SESSINFO_SERVER_PORT 2000
    cryptSetAttribute $session CRYPT_SESSINFO_PRIVATEKEY \
                                    [lindex $tempKeyPairs 0]
    cryptSetAttribute $session CRYPT_SESSINFO_ACTIVE 1
    puts "Connection established..."
    cryptSetAttribute CRYPT_UNUSED CRYPT_OPTION_NET_READTIMEOUT 30
    cryptSetAttribute CRYPT_UNUSED CRYPT_OPTION_NET_WRITETIMEOUT 30
    set text [binary format a10240 \0]
    set bytesSent 0
    set bytesRead 0
    for {set count 0} {$count < 512} {incr count} {
        set bytesRead 0
        while {$bytesRead < 10240} {
            cryptPopData $session text 10240 bytesCopied
            incr bytesRead $bytesCopied
        }
        set bytesSent 0
        while {$bytesSent < 10240} {
            cryptPushData $session $text bytesCopied
            incr bytesSent $bytesCopied
        }
        cryptFlushData $session
    }
    cryptDestroySession $session
    puts "Session closed..."
}

proc testCMS {} {
    global tempKeyPairs tempCerts tempKeySet
    set plaintext "Hi, how are you"
    set bufsize [expr {31 * 1024}]
    set buffer [binary format a$bufsize \0]

    # First sign with temp key 0
    puts "Signing message"
    cryptCreateEnvelope e CRYPT_UNUSED CRYPT_FORMAT_CRYPTLIB
    cryptSetAttribute $e CRYPT_ENVINFO_SIGNATURE [lindex $tempKeyPairs 0]
    cryptPushData $e $plaintext bytesCopied
    cryptFlushData $e
    cryptPopData $e buffer $bufsize bytesCopied
    cryptDestroyEnvelope $e

    # Then encrypt to temp certs 0 and 1, using AES256
    puts "Encrypting message"
    cryptCreateEnvelope e CRYPT_UNUSED CRYPT_FORMAT_CRYPTLIB

    cryptCreateContext c CRYPT_UNUSED CRYPT_ALGO_AES
    cryptSetAttribute $c CRYPT_CTXINFO_KEYSIZE 32
    cryptGenerateKey $c
    cryptSetAttribute $e CRYPT_ENVINFO_SESSIONKEY $c
    cryptDestroyContext $c

    cryptSetAttribute $e CRYPT_ENVINFO_PUBLICKEY [lindex $tempCerts 0]
    cryptSetAttribute $e CRYPT_ENVINFO_PUBLICKEY [lindex $tempCerts 1]

    cryptPushData $e $buffer bytesCopied
    cryptFlushData $e
    cryptPopData $e buffer $bufsize bytesCopied
    cryptDestroyEnvelope $e

    # prepare to decrypt encrypted data
    puts "Decrypting message"
    cryptCreateEnvelope e CRYPT_UNUSED CRYPT_FORMAT_AUTO

    # It tells us we need to an encryption attribute, via Exceptions
    # This is an unfortunate result of translating status values as
    # Exceptions, but we can live with it
    if {[catch {cryptPushData $e $buffer bytesCopied} status] \
         && $status ne "CRYPT_ENVELOPE_RESOURCE"} {
            error "Cryptlib error - $status"
    }
    if {[catch {cryptFlushData $e} status] \
         && $status ne "CRYPT_ENVELOPE_RESOURCE"} {
            error "Cryptlib error - $status"
    }

    # Iterate through encryption attributes until we find one that matches key 1
    cryptSetAttribute $e CRYPT_ATTRIBUTE_CURRENT CRYPT_CURSOR_FIRST
    while {1} {
        cryptGetAttribute $e CRYPT_ATTRIBUTE_CURRENT attributeNeeded
        if {$attributeNeeded == $::cryptkit::CRYPT_ENVINFO_PRIVATEKEY} {
            # Test if key 1 works. If so, exit the loop
            catch {cryptSetAttribute $e CRYPT_ENVINFO_PRIVATEKEY \
                            [lindex $tempKeyPairs 1]} status
            if {$status eq "CRYPT_OK"} {
                break
            }
            # This is not the correct key
            if {$status ne "CRYPT_ERROR_WRONGKEY"} {
                error "Cryptlib error - $status"
            }
        }
        catch {
            cryptSetAttribute $e CRYPT_ATTRIBUTE_CURRENT CRYPT_CURSOR_NEXT
        } status
        if {$status eq "CRYPT_ERROR_NOTFOUND"} break
    }
    cryptPopData $e buffer $bufsize bytesCopied
    # note sure why, but in Cryptlib 3.1 destroying $e at this point causes a
    # CRYPT_ERROR_INCOMPLETE error
    cryptDestroyEnvelope $e

    # prepare to verify decrypted data
    puts "Veryifying message"
    cryptCreateEnvelope e CRYPT_UNUSED CRYPT_FORMAT_AUTO

    # Check that it was signed by one of the keys in tempKeySet
    cryptSetAttribute $e CRYPT_ENVINFO_KEYSET_SIGCHECK $tempKeySet

    cryptPushData $e $buffer bytesCopied
    cryptFlushData $e
    cryptPopData $e buffer $bufsize bytesCopied

    cryptGetAttribute $e CRYPT_ENVINFO_SIGNATURE_RESULT result
    if {$result != $::cryptkit::CRYPT_OK} {
        error "signature failed"
    }
    cryptDestroyEnvelope $e

    puts "Decrypted/Verified data is:\n[string range $buffer 0 [incr bytesCopied -1]]"
}

proc testCipher {cryptAlgo keySize algoName} {
    # Create context with random IV (if it needs one) and a key derived
    # from a password using a salt and iteration count to hinder guessing
    # attacks
    cryptCreateContext encContext CRYPT_UNUSED $cryptAlgo
    cryptSetAttribute $encContext CRYPT_CTXINFO_KEYSIZE $keySize
    cryptSetAttribute $encContext CRYPT_CTXINFO_KEYING_ITERATIONS 10000
    cryptSetAttributeString $encContext CRYPT_CTXINFO_KEYING_SALT "salt1234"
    cryptSetAttributeString $encContext CRYPT_CTXINFO_KEYING_VALUE "password123"

    # Encrypt 10 MB of text
    # We do this in 10 KB chunks for the sake of demonstration

    set text [binary format a[expr {10 * 1024 * 1024}] \0]
    set startTime [clock seconds]
    
    cryptEncrypt $encContext $text
    set endTime [clock seconds]
    puts "$cryptAlgo 10 Mb in [expr {$endTime - $startTime}] seconds"

    cryptDestroyContext $encContext
}

proc testBlobs {} {
    # Demonstrate cryptQueryCapability()
    cryptQueryCapability CRYPT_ALGO_AES q
    puts "AES capabilities:"
    puts "algoName = $q(algoName)"
    puts "blockSize = $q(blockSize)"
    puts "minKeySize = $q(minKeySize)"
    puts "keySize = $q(keySize)"
    puts "maxKeySize = $q(maxKeySize)"

    # Hash some data
    cryptCreateContext hashContext CRYPT_UNUSED CRYPT_ALGO_SHA
    cryptEncrypt $hashContext "hello I am data to hash"
    cryptEncrypt $hashContext ""

    # Generate DSA key
    cryptCreateContext dsaContext CRYPT_UNUSED CRYPT_ALGO_DSA
    cryptSetAttributeString $dsaContext CRYPT_CTXINFO_LABEL "testkey"
    cryptGenerateKey $dsaContext

    # Produce a signature
    set buffer [binary format a1000 \0]
    cryptCreateSignature buffer siglength $dsaContext $hashContext 

    # Query the signature blob
    cryptQueryObject $buffer i
    puts "Signature Properties: "
    puts "objectType = $i(objectType)"
    puts "cryptAlgo = $i(cryptAlgo)"
    puts "cryptMode = $i(cryptMode)"
    puts "hashAlgo = $i(hashAlgo)"
    puts "salt = $i(salt)"

    # Validate the signature
    cryptCheckSignature $buffer $dsaContext $hashContext
    puts "The signature validates correctly!"

    # Generate a symmetric key
    cryptCreateContext aesContext CRYPT_UNUSED CRYPT_ALGO_AES
    cryptGenerateKey $aesContext

    # Generate a symmetric key to wrap the first one
    cryptCreateContext desContext CRYPT_UNUSED CRYPT_ALGO_3DES
    cryptSetAttributeString $desContext CRYPT_CTXINFO_KEYING_SALT \
                        "this is salt sdfsdf"
    cryptSetAttributeString $desContext CRYPT_CTXINFO_KEYING_VALUE \
                        "secret password"

    # Wrap the first key in the second
    cryptExportKey buffer exportedKeyLength $desContext $aesContext

    # Query the encrypted key
    cryptQueryObject $buffer i
    puts "Exported Key Properties: "
    puts "objectType = $i(objectType)"
    puts "cryptAlgo = $i(cryptAlgo)"
    puts "cryptMode = $i(cryptMode)"
    puts "hashAlgo = $i(hashAlgo)"
    puts "salt = $i(salt)"

    cryptDestroyContext $dsaContext
    cryptDestroyContext $hashContext
    cryptDestroyContext $aesContext
    cryptDestroyContext $desContext
}
